/*
 * Simple but hopefully safer RPC service replacement
 *
 * Copright (C) 2000 <okir@caldera.de>
 */

#include <sys/time.h>
#include <malloc.h>
#include <unistd.h>
#include <errno.h>
#include "auth_unix.h"
#include "svc_int.h"

#define SVC_TIMEOUT	30
#define SVC_MAXREQSIZE	(16 * 1024)

/*
 * Local variables
 */
static SVCXPRT *	all_ports = NULL;
static unsigned int	portlist_changed = 0;
static unsigned int	nsocks = 0;
static int		exit_loop = 0;

/*
 * Internal functions
 */
static bool_t		svc_authenticate(SVCXPRT *, struct svc_req *);
static void		xprt_free(SVCXPRT *p);
static int		xprt_makesock(int proto);

#ifndef MAX
# define MAX(a, b)	(((a) > (b))? (a) : (b))
#endif

/*
 * Register a service on a given transport.
 */
bool_t
svc_register(SVCXPRT *xprt, u_int32_t prog, u_int32_t vers,
		__dispatch_fn_t handler, int protocol)
{
	SERVICE	*svc;

	/* Not permitted to register service on an established TCP
	 * connection -- always register with the listening socket */
	if (xprt->parent)
		return 0;

	svc = (SERVICE *) malloc(sizeof(*svc));
	if (svc == NULL)
		return 0;
	
	memset(svc, 0, sizeof(*svc));
	svc->handler = handler;
	svc->prog = prog;
	svc->vers = vers;

	/* Register with portmapper */
	if (protocol) {
		/* XXX to be done */
	}

	/* Add this service to the transport's list of services */
	svc->next = xprt->service;
	xprt->service = svc;

	return 1;
}

void
svc_unregister(u_int32_t prog, u_int32_t vers)
{
	SVCXPRT	*p;
	SERVICE	*s, **ss;

again:
	for (p = all_ports; p; p = p->next) {
		ss = &p->service;
		while ((s = *ss) != NULL) {
			if (s->prog == prog && s->vers == vers) {
				*ss = s->next;
				free(s);
			} else {
				ss = &s->next;
			}
		}
		if (p->service == NULL) {
			svc_destroy(p);
			goto again;
		}
	}
}

/*
 * Main loop
 */
void
svc_run()
{
	while (all_ports && !exit_loop) {
		struct timeval	when, tv, now, *tvp;
		SVCXPRT		*p;
		fd_set		rset, wset;
		int		n, maxfd = -1;

		timerclear(&when);
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		for (p = all_ports; p; p = p->next) {
			FD_SET(p->fd, (p->receiving? &rset : &wset));
			if (p->fd > maxfd)
				maxfd = p->fd;

			/* Handle timeout */
			if (!timerisset(&when)
			 || (timerisset(&p->timeout)
			 && timercmp(&p->timeout, &when, <)))
				when = p->timeout;
		}

		/* timeout handling */
		tvp = NULL;
		if (timerisset(&when)) {
			timerclear(&tv);
			gettimeofday(&now, NULL);
			if (timercmp(&when, &now, >=))
				timersub(&when, &now, &tv);
			tvp = &tv;
		}

		n = select(maxfd+1, &rset, &wset, NULL, tvp);

		/* Loop over all transport to see if any of them is
		 * able to recv/xmit data.
		 * Any of the handlers may delete the transport
		 * it's working on (e.g. due to a socket error).
		 * In this case, the port list is altered, and we
		 * need to start from the beginning so we don't
		 * trip over a stale pointer.
		 */
		gettimeofday(&now, NULL);
		do {
			portlist_changed = 0;
			for (p = all_ports; p; p = p->next) {
				if (FD_ISSET(p->fd, &rset)) {
					FD_CLR(p->fd, &rset);
					p->receiver(p);
				} else
				if (FD_ISSET(p->fd, &wset)) {
					FD_CLR(p->fd, &wset);
					p->sender(p);
				}
				else
				if (timercmp(&now, &p->timeout, <=)) {
					svc_destroy(p);
				}

				if (portlist_changed)
					break;
			}
		} while (portlist_changed);
	}
	portlist_changed = 1;
	exit_loop = 0;
}

/*
 * Exit the service loop
 */
void
svc_exit()
{
	exit_loop = 1;
}

/*
 * Close a transport.
 * If it's a TCP transport, also close all connections
 * accepted on this socket
 */
void
svc_destroy(SVCXPRT *p)
{
	SVCXPRT	**pp, *q;
	int		dead = 0;

	if (p == NULL)
		return;

	pp = &all_ports;
	while ((q = *pp) != NULL) {
		if (q == p || q->parent == p) {
			if (p == q)
				dead = 1;
			portlist_changed = 1;
			*pp = q->next;
			xprt_free(q);
			nsocks--;
		} else {
			pp = &q->next;
		}
	}

	if (!dead)
		xprt_free(p);
}

/*
 * Start idle timer
 */
static void
svc_settimer(SVCXPRT *p)
{
	gettimeofday(&p->timeout, NULL);
	p->timeout.tv_sec += SVC_TIMEOUT;
}

/*
 * Execute a request
 */
static void
svc_doit(SVCXPRT *p)
{
	struct svc_req	req;
	XDR		*xdr;
	struct rpc_msg	call, *resp;
	SERVICE		*s;
	u_int32_t	minvers, maxvers;

	/* We've received a complete request.
	 * Nuke the timer */
	timerclear(&p->timeout);

	/* Initialize the XDR codecs for the receive and send buffers */
	xdr = &p->rinfo.xdrs;
	xdrmem_create(xdr, p->rinfo.buffer, p->rinfo.bufsize, XDR_DECODE);

	/* Parse the call header */
	memset(&call, 0, sizeof(call));
	if (!xdr_callhdr(xdr, &call))
		return;

	resp = &p->sinfo.resp;
	memset(resp, 0, sizeof(*resp));
	resp->rm_xid = call.rm_xid;
	resp->rm_direction = REPLY;
	if (call.rm_call.cb_rpcvers != 2) {
		resp->rm_reply.rp_stat = MSG_DENIED;
		resp->rm_reply.rp_rjct.rj_stat = RPC_MISMATCH;
		resp->rm_reply.rp_rjct.rj_vers.low = 2;
		resp->rm_reply.rp_rjct.rj_vers.high = 2;
		svc_sendreply(p, NULL, NULL);
		goto done;
	}

	/* Set up the svc_req */
	memset(&req, 0, sizeof(req));
	req.rq_xprt = p;
	req.rq_prog = call.rm_call.cb_prog;
	req.rq_vers = call.rm_call.cb_vers;
	req.rq_proc = call.rm_call.cb_proc;
	memcpy(&req.rq_cred, &call.rm_call.cb_cred, sizeof(req.rq_cred));
	memcpy(&req.rq_verf, &call.rm_call.cb_verf, sizeof(req.rq_verf));
	if (!svc_authenticate(p, &req)) {
		svc_sendreply(p, NULL, NULL);
		goto done;
	}

	resp->rm_reply.rp_stat = MSG_ACCEPTED;

	/* Find the service that implements this program/version */
	s = (p->parent? p->parent : p)->service;
	maxvers = 0; minvers = 0xFFFFFFFF;
	for (; s; s = s->next) {
		if (s->prog != call.rm_call.cb_prog)
			continue;
		if (s->vers == call.rm_call.cb_vers)
			break;
		if (s->vers > maxvers)
			maxvers = s->vers;
		if (s->vers < minvers)
			minvers = s->vers;
	}
	if (s != NULL) {
		/* Amazing, we actually serve this program.
		 * Call the handler */
		s->handler(&req, p);
	} else {
		/* Too bad, no cookies. Decide on the proper
		 * error message (bad program/bad version)... */
		if (minvers > maxvers)
			svcerr_noprog(p);
		else
			svcerr_progvers(p, minvers, maxvers);
	}

done:
	/* Done with the request. If a response has been generated
	 * and is pending transmission, p->receiving will be zero.
	 * Otherwise, we restart the receiver */
	if (p->receiving && p->startrecv)
		p->startrecv(p);
}

/*
 * Perform authentication
 */
static bool_t
svc_authenticate(SVCXPRT *p, struct svc_req *rqstp)
{
	static struct authunix_parms aup; /* FIXME: not thread safe */
	struct opaque_auth *cred = &rqstp->rq_cred;
	struct rpc_msg *resp = &p->sinfo.resp;
	XDR		xdr;

	/* For now, we don't do much authentication. We accept
	 * NONE and UNIX, and that's it. */

	rqstp->rq_clntcred = NULL;
	switch (cred->oa_flavor) {
	case AUTH_NONE:
		break;

	case AUTH_UNIX:
		xdrmem_create(&xdr, cred->oa_base, cred->oa_length, XDR_DECODE);
		if (xdr_authunix_parms(&xdr, &aup)) {
			rqstp->rq_clntcred = &aup;
			break;
		}

		/* fallthrough */
	default:
		svcerr_auth(p, AUTH_BADCRED);
		return 0;
	}

	/* Mark as okay, add NULL verifier */
	resp->rm_reply.rp_stat = MSG_ACCEPTED;
	resp->rm_reply.rp_acpt.ar_stat = SUCCESS;
	resp->rm_reply.rp_acpt.ar_verf.oa_flavor = AUTH_NONE;
	resp->rm_reply.rp_acpt.ar_verf.oa_base = NULL;
	resp->rm_reply.rp_acpt.ar_verf.oa_length = 0;
	return 1;
}

/*
 * Error functions
 */
void
svcerr_auth(SVCXPRT *p, enum auth_stat why)
{
	struct rpc_msg *resp = &p->sinfo.resp;

	resp->rm_reply.rp_stat = MSG_DENIED;
	resp->rm_reply.rp_rjct.rj_stat = AUTH_ERROR;
	resp->rm_reply.rp_rjct.rj_why  = why;
	svc_sendreply(p, NULL, NULL);
}

static inline void
svc_senderror(SVCXPRT *p, int why)
{
	struct rpc_msg *resp = &p->sinfo.resp;

	resp->rm_reply.rp_stat = MSG_ACCEPTED;
	resp->rm_reply.rp_acpt.ar_stat = why;
	svc_sendreply(p, NULL, NULL);
}

void
svcerr_decode(SVCXPRT *p)
{
	svc_senderror(p, GARBAGE_ARGS);
}

void
svcerr_weakauth(SVCXPRT *p)
{
	svcerr_auth(p, AUTH_TOOWEAK);
}

void
svcerr_noprog(SVCXPRT *p)
{
	svc_senderror(p, PROG_UNAVAIL);
}

void
svcerr_progvers(SVCXPRT *p, u_int32_t low, u_int32_t high)
{
	p->sinfo.resp.rm_reply.rp_acpt.ar_vers.low = low;
	p->sinfo.resp.rm_reply.rp_acpt.ar_vers.high = high;
	svc_senderror(p, PROG_MISMATCH);
}

void
svcerr_noproc(SVCXPRT *p)
{
	svc_senderror(p, PROC_UNAVAIL);
}

void
svcerr_systemerr(SVCXPRT *p)
{
	svc_senderror(p, SYSTEM_ERR);
}

/*
 * Get the caller's address
 */
struct sockaddr *
svc_getcaller(SVCXPRT *p)
{
	return (struct sockaddr *) &p->peer;
}

struct sockaddr *
svc_getsvcaddr(SVCXPRT *p)
{
	return (struct sockaddr *) &p->addr;
}

/* 
 * Parse the arguments.
 * This is easy... we've already parsed the header, now go for
 * the body.
 */
bool_t
svc_getargs(SVCXPRT *p, xdrproc_t decode, void *ptr)
{
	return decode(&p->rinfo.xdrs, ptr);
}

/*
 * Construct a response in the output buffer
 */
bool_t
svc_sendreply(SVCXPRT *p, xdrproc_t encode, void *ptr)
{
	XDR	xdrs;

	/* Called twice from server? This is a bug */
	if (!p->receiving)
		return 0;

	/* Create dynamic memory XDR handler */
	xdrmem_create(&xdrs, p->sinfo.buffer, p->sinfo.bufsize, XDR_ENCODE);
	if (encode) {
		p->sinfo.resp.rm_reply.rp_acpt.ar_results.proc = encode;
		p->sinfo.resp.rm_reply.rp_acpt.ar_results.where = ptr;
	}
	if (!xdr_replymsg(&xdrs, &p->sinfo.resp))
		return 0;

	p->startxmit(p, &xdrs);
	return 1;
}

/*
 * Create an empty SVCXPRT
 */
static SVCXPRT *
xprt_new(int fd, unsigned int rcvbuf, unsigned int sndbuf)
{
	SVCXPRT		*p;
	socklen_t	alen;

	p = (SVCXPRT *) malloc(sizeof(*p));
	if (p == NULL) {
		close(fd);
		return NULL;
	}
	memset(p, 0, sizeof(*p));
	p->fd = fd;
	p->rinfo.bufsize = MAX(rcvbuf, SVC_MAXREQSIZE);
	p->rinfo.buffer  = malloc(p->rinfo.bufsize);
	p->sinfo.bufsize = MAX(sndbuf, SVC_MAXREQSIZE);
	p->sinfo.buffer  = malloc(p->sinfo.bufsize);
	if (p->rinfo.buffer == NULL || p->sinfo.buffer == NULL) {
		xprt_free(p);
		return NULL;
	}

	alen = sizeof(p->addr);
	if (getsockname(fd, (struct sockaddr *) &p->addr, &alen) < 0) {
		xprt_free(p);
		return NULL;
	}

	/* Link into list of all transports */
	p->next = all_ports;
	all_ports = p;
	nsocks++;

	return p;
}

static void
xprt_free(SVCXPRT *p)
{
	SERVICE	*s;

	if (p->rinfo.buffer)
		free(p->rinfo.buffer);
	if (p->sinfo.buffer)
		free(p->sinfo.buffer);
	while ((s = p->service) != NULL) {
		p->service = s->next;
		free(s);
	}
	if (p->fd >= 0)
		close(p->fd);
	free(p);
}

/*
 * UDP transport
 */
static void
udp_receiver(SVCXPRT *p)
{
	struct svc_rinfo *r = &p->rinfo;
	int		n;
	socklen_t	salen;

	salen = sizeof(p->peer);
	n = recvfrom(p->fd, r->buffer, r->bufsize, 0,
			(struct sockaddr *) &p->peer, &salen);
	if (n <= 0)
		return;
	r->size = r->have = n;
	svc_doit(p);
}

static void
udp_startxmit(SVCXPRT *p, XDR *xdrs)
{
	sendto(p->fd, ((caddr_t) xdrs->x_base), xdrs->x_pos, 0,
			(struct sockaddr *) &p->peer, sizeof(p->peer));
}

SVCXPRT *
svcudp_bufcreate(int sock, unsigned int rcvbuf, unsigned int sndbuf)
{
	SVCXPRT	*p;

	/* Check if wildcard socket given, and make one if so */
	if (sock == RPC_ANYSOCK && (sock = xprt_makesock(SOCK_DGRAM)) < 0)
		return NULL;

	if ((p = xprt_new(sock, rcvbuf, sndbuf)) == NULL)
		return NULL;
	p->receiving = 1;
	p->receiver  = udp_receiver;
	p->startxmit = udp_startxmit;
	return p;
}

SVCXPRT *
svcudp_create(int sock)
{
	return svcudp_bufcreate(sock, 0, 0);
}

/*
 * Create a TCP service socket
 */
static inline void
tcp_nextfrag(SVCXPRT *p)
{
	p->rinfo.startofrecord = 1;
	p->rinfo.want = 4;
	p->rinfo.have = 0;
}

static void
tcp_startrecv(SVCXPRT *p)
{
	svc_settimer(p);
	p->receiving = 1;
	p->rinfo.size = 0;
	tcp_nextfrag(p);
}

static void
tcp_receiver(SVCXPRT *p)
{
	struct svc_rinfo *r = &p->rinfo;
	u_int32_t	marker, oldsize;
	int		n;

	n = read(p->fd, r->buffer + r->have, r->want);
	if (n <= 0) {
		svc_destroy(p);
		return;
	}

	r->have += n;
	r->want -= n;
	if (r->want != 0)
		return;
	
	/* We have read all we wanted to (for now).
	 * Check what we have got: either a 4 byte record
	 * marker, or a record (fragment).
	 */
	if (!r->startofrecord) {
		/* We've read a record (fragment). 
		 * Are there more to come, or can we execute
		 * the request straightaway? */
		if (r->morefrags) {
			tcp_nextfrag(p);
		} else {
			svc_doit(p);
		}
		return;
	}

	/* We have a 4 byte record marker. Dissect it and
	 * see what's coming.
	 */
	r->startofrecord = r->morefrags = 0;
	r->have -= 4;
	memcpy(&marker, r->buffer + r->have, 4);
	r->want = ntohl(marker) & 0x7FFFFFFF;
	if (marker & 0x80000000)
		r->morefrags = 1;

	/* Sanity checks */
	oldsize = r->size;
	r->size += r->want;
	if (r->size <= oldsize		/* int overflow or want == 0 */
	 || r->size >= r->bufsize) {
		svc_destroy(p);
		return;
	}
}

static void
tcp_sender(SVCXPRT *p)
{
	struct svc_sinfo *s = &p->sinfo;
	int		n;

	if (s->startofrecord) {
		u_int32_t	marker = htonl(s->size);
		u_int8_t	*data;

		data = (u_int8_t *) &marker;
		n = write(p->fd, data + s->sent, 4 - s->sent);
	} else {
		n = write(p->fd, s->buffer + s->sent, s->size - s->sent);
	}

	if (n <= 0) {
		svc_destroy(p);
		return;
	}
	s->sent += n;

	if (s->startofrecord) {
		if (s->sent == 4) {
			s->startofrecord = 0;
			s->sent = 0;
		}
	}
	if (s->sent == s->size) {
		p->receiving = 1;
		free(s->buffer);
		s->buffer = NULL;
	}
}

static void
tcp_startxmit(SVCXPRT *p, XDR *xdrs)
{
	svc_settimer(p);
	p->sinfo.startofrecord = 1;
	p->sinfo.sent = 0;
	p->sinfo.size = xdrs->x_pos;
	p->receiving = 0;
}

static void
tcp_rendezvouser(SVCXPRT *p)
{
	SVCXPRT		*pnew;
	struct sockaddr_in peer;
	socklen_t	salen;
	int		fd;

	salen = sizeof(p->peer);
	fd = accept(p->fd, (struct sockaddr *) &peer, &salen);
	if (fd < 0)
		return;

	/* When we're above a certain threshold, reject connects
	 * from anyone but local root */
#ifdef RPC_SOCKTHRESH
	if (nsocks >= RPC_SOCKTHRESH
	 && (ntohs(p->peer.sin_port) < 512
	  || ntohs(p->peer.sin_port) > 1023
	  || ntohl(p->peer.sin_addr.s_addr) != INADDR_ANY)) {
		close(fd);
		return;
	}
#endif

	pnew = xprt_new(fd, p->rinfo.bufsize, 0);
	if (pnew == NULL)
		return;
	pnew->parent    = p;
	pnew->peer      = peer;
	pnew->startrecv = tcp_startrecv;
	pnew->startxmit = tcp_startxmit;
	pnew->receiver  = tcp_receiver;
	pnew->sender    = tcp_sender;

	pnew->startrecv(pnew);
}

SVCXPRT *
svctcp_create(int sock, unsigned int rcvbuf, unsigned int sndbuf)
{
	SVCXPRT	*p;

	/* Check if wildcard socket given, and make one if so */
	if (sock == RPC_ANYSOCK && (sock = xprt_makesock(SOCK_STREAM)) < 0)
		return NULL;

	if ((p = xprt_new(sock, rcvbuf, sndbuf)) == NULL)
		return NULL;
	if (listen(p->fd, 128) < 0) {
		svc_destroy(p);
		return NULL;
	}

	p->receiving = 1;
	p->receiver  = tcp_rendezvouser;
	return p;
}

/*
 * Create a socket
 */
static int
xprt_makesock(int type)
{
	struct sockaddr_in sin;
	static int	port = 666;
	int		fd;

	if ((fd = socket(PF_INET, type, 0)) < 0)
		return -1;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;

	while (1) {
		sin.sin_port = htons(port);
		if (bind(fd, (struct sockaddr *) &sin, sizeof(sin)) >= 0)
			return fd;

		if (port == 0)
			break;
		if (errno != EADDRINUSE)
			port = 0;
		else if (++port >= 1024)
			break;
	}

	close(fd);
	return -1;
}
