/*
 * Simple but hopefully safer RPC service replacement
 *
 * Copright (C) 2000 <okir@caldera.de>
 *
 * This implementation tries to protect against the most common
 * DoS attacks against TCP based RPC servers.
 *
 *  1.	Send a record marker, then stop. This will hang the server
 *	for 30 secs or so.
 *  2.	Send a valid RPC packet, byte by byte, 20 seconds apart.
 *	This will hang the server for as long as it takes to transmit
 *	the packet.
 *  3.	Send a stream of valid packets to the server, but never
 *	read the replies from your TCP socket. This will gradually
 *	fill up the network pipe until the server blocks in tcp_write.
 *	This blocking is indefinite.
 *
 * What this implementation doesn't do:
 *  1.	It supports only NONE and UNIX credentials. No DES or GSS
 *  2.	It has an upper limit on the maximum size of requests it
 *	will handle (16K currently). Useful for most system services,
 *	but I've seen apps that shove several megs worth of data through
 *	a TCP stream. Ick.
 *  3.	Protect you against "eat up all the server's TCP sockets"
 *	type of DoS attacks.
 */

typedef struct _srvc {
	struct _srvc *	next;
	u_int32_t	prog;
	u_int32_t	vers;
	void		(*handler)(struct svc_req *, SVCXPRT *);
} SERVICE;

typedef struct _svcxprt {
	struct _xprt *	next;
	int		fd;
	SERVICE *	service;
	unsigned int	running   : 1,
			receiving : 1;

	struct svc_rinfo {
		unsigned int	startofrecord : 1,
				morefrags     : 1;
		u_int8_t *	buffer;
		u_int32_t	bufsize;
		u_int32_t	want;
		u_int32_t	have;
		u_int32_t	size;
		XDR		xdrs;
	}		rinfo;
	struct svc_sinfo {
		unsigned int	startofrecord : 1;
		u_int8_t *	buffer;
		u_int32_t	sent;
		u_int32_t	size;
	}		sinfo;

	struct sockaddr_in peer;

	void		(*startrecv)(struct port *);
	void		(*receiver)(struct port *);
	void		(*startxmit)(struct port *);
	void		(*sender)(struct port *);
} SVCXPRT;

#define RPC_MAXBUFSIZE	(16 * 1024)

/*
 * Local variables
 */
static SVCXPRT *	all_ports = NULL;
static unsigned int	portlist_changed = 0;
static unsigned int	nsocks = 0;

/*
 * Main loop
 */
static void
svc_run()
{
	while (all_ports) {
		fd_set	rset, wset;
		int	maxfd = -1;

		FD_ZERO(&rset);
		FD_ZERO(&wset);
		for (p = all_ports; p; p = p->next) {
			FD_SET(p->fd, (p->receiving? &rset : &wset));
			if (p->fd > maxfd)
				maxfd = p->fd;
		}

		/* XXX: timeout handling */

		n = select(maxfd+1, &rset, &wset, NULL, NULL);

		/* Loop over all transport to see if any of them is
		 * able to recv/xmit data.
		 * Any of the handlers may delete the transport
		 * it's working on (e.g. due to a socket error).
		 * In this case, the port list is altered, and we
		 * need to start from the beginning so we don't
		 * trip over a stale pointer.
		 */
		do {
			portlist_changed = 0;
			for (p = all_ports; p; p = p->next) {
				if (FD_ISSET(p->fd, &rset))) {
					FD_CLR(p->fd, &rset);
					p->receiver(p);
				} else
				if (FD_ISSET(p->fd, &wset))) {
					FD_CLR(p->fd, &wset);
					p->sender(p);
				}
				if (portlist_changed)
					break;
			}
		} while (p != NULL);
	}
}

/*
 * Register a service on a given transport.
 */
static int
svc_register(SVCXPRT *xprt, u_int32_t prog, u_int32_t vers,
		void (*handler)(struct svc_req *, SVCXPRT *),
		int funny)
{
	SERVICE	*svc;

	/* Not permitted to register service on an established TCP
	 * connection -- always register with the listening socket */
	if (xprt->parent)
		return NULL;

	svc = (SERVICE *) malloc(sizeof(*svc));
	if (svc == NULL)
		return 0;
	
	memset(svc, 0, sizeof(*svc));
	svc->handler = handler;
	svc->prog = prog;
	svc->vers = vers;

	/* Add this service to the transport's list of services */
	svc->next = xprt->service;
	xprt->service = svc;

	return 1;
}

/*
 * Close a transport.
 * If it's a TCP transport, also close all connections
 * accepted on this socket
 */
static void
svc_destroy(SVCXPRT *p)
{
	SVCXPRT	**pp, *q;
	int		dead = 0;

	if (p == NULL)
		return;

	pp = &all_ports;
	while ((q = *pp) != NULL) {
		if (q == p || q->parent == p) {
			if (p == q)
				dead = 1;
			portlist_changed = 1;
			*pp = q->next;
			xprt_free(q);
			nsocks--;
		} else {
			pp = &q->next;
		}
	}

	if (!dead)
		xprt_free(p);
}

/*
 * Execute a request
 */
static void
svc_doit(SVCXPRT *p)
{
	struct svc_req	req;
	XDR		*xdr;
	struct xdr_msg	call, resp;
	SERVICE		*s;
	u_int32_t	minvers, maxvers;

	/* Initialize the XDR codecs for the receive and send buffers */
	xdr = &p->rinfo.xdrs;
	xdrmem_create(xdr, p->rinfo.buffer, p->rinfo.size);

	/* Parse the call header */

	/* Set up the svc_req */

	/* Find the service that implements this program/version */
	s = (p->parent? p->parent : p)->service;
	maxvers = 0; minvers = 0xFFFFFFFF;
	for (; s; s = s->next) {
		if (s->prog != call.rm_call.cb_prog)
			continue;
		if (s->vers == call.rm_call.cb_vers)
			break;
		if (s->vers > maxvers)
			maxvers = s->vers;
		if (s->vers < minvers)
			minvers = s->vers;
	}
	if (s != NULL) {
		/* Amazing, we actually serve this program.
		 * Call the handler */
		s->handler(&req, p);
	} else {
		/* Too bad, no cookies. Decide on the proper
		 * error message (bad program/bad version)... */
		if (minvers > maxvers)
			svcerr_noprog(p);
		else
			svcerr_novers(p, minvers, maxvers);
	}

done:
	/* Done with the request. If a response has been generated
	 * and is pending transmission, p->receiving will be zero.
	 * Otherwise, we restart the receiver */
	if (p->receiving && p->startrecv)
		p->startrecv(p);
}

/* 
 * Parse the arguments.
 * This is easy... we've already parsed the header, now go for
 * the body.
 */
int
svc_getargs(SVCXPRT *p, xdrfunc_t decode, caddr_t ptr)
{
	return decode(&p->rinfo.xdrs, ptr, XDR_DECODE);
}

/*
 * Construct a response in the output buffer
 */
int
svc_sendreply(SVCXPRT *p, xdrfunc_t encode, caddr_t ptr)
{
	XDR	xdrs;

	/* Called twice from server? This is a bug */
	if (!p->receiving)
		return;

	/* Create dynamic memory XDR handler */
	xdr_dmemcreate(&xdrs);
	if (!encode(&xdrs, ptr, XDR_ENCODE)) {
		xdr_destroy(&xdrs);
		return;
	}

	p->startxmit(p, &xdrs);
}

/*
 * Create an empty SVCXPRT
 */
static SVCXPRT *
xprt_new(int fd, unsigned int rcvbuf, unsigned int sndbuf)
{
	SVCXPRT	*p;

	/* Check if wildcard socket given, and make one if so */
	if (fd == RPC_ANYSOCK && (fd = xprt_makesock()) < 0)
		return NULL;

	p = (SVCXPRT *) malloc(sizeof(*p));
	if (p == NULL) {
		close(fd);
		return NULL;
	}
	memset(p, 0, sizeof(*p));
	p->fd = fd;
	p->rinfo.bufsize = MAX(rcvbuf, RPC_MAXREQSIZE);
	p->rinfo.buffer  = (u_int8_t *) malloc(p->rinfo.bufsize);
	if (p->rinfo.buffer == NULL) {
		xprt_free(p);
		return NULL;
	}

	/* Link into list of all transports */
	p->next = all_ports;
	all_ports = p;
	nsocks++;

	return p;
}

static void
xprt_free(SVCXPRT *p)
{
	SERVICE	*s;

	if (p->rinfo.buffer)
		free(p->rinfo.buffer);
	if (p->sinfo.buffer)
		free(p->sinfo.buffer);
	while ((s = p->service) != NULL) {
		p->service = s->next;
		free(s);
	}
	if (p->fd >= 0)
		close(p->fd);
	free(p);
}

/*
 * UDP transport
 */
SVCXPRT *
svcudp_create(int sock)
{
	SVCXPRT	*p;

	if ((p = xprt_new(sock, 0, 0)) == NULL)
		return NULL;
	p->receiving = 1;
	p->receiver  = udp_receiver;
	p->startxmit = udp_startxmit;
	return p;
}

static void
udp_receiver(SVCXPRT *p)
{
	struct svc_rinfo *r = &p->rinfo;
	int		n;
	socklen_t	salen;

	salen = sizeof(p->peer);
	n = recvfrom(p->fd, r->buffer, r->bufsize, 0,
			(struct sockaddr *) &p->peer, &salen);
	if (n <= 0)
		return;
	r->size = r->have = n;
	svc_doit(p);
}

static void
udp_startxmit(SVCXPRT *p, XDR *xdrs)
{
	sendto(p->fd, xdrs->buffer, xdrs->size, 0,
			(struct sockaddr *) &p->peer, sizeof(p->peer));
}

/*
 * Create a TCP service socket
 */
SVCXPRT *
svc_tcpcreate(int sock, unsigned int rcvbuf, unsigned int sndbuf)
{
	SVCXPRT	*p;

	if ((p = xprt_new(sock, rcvbuf, sndbuf)) == NULL)
		return NULL;
	if (listen(p->fd, 128) < 0) {
		svc_destroy(p);
		return NULL;
	}

	p->receiving = 1;
	p->receiver  = tcp_rendezvouser;
	return p;
}

static void
tcp_rendezvouser(SVCXPRT *p)
{
	SVCXPRT		*pnew;
	SERVICE		*svc;
	struct sockaddr_in peer;
	socklen_t	salen;
	int		fd;

	salen = sizeof(p->peer);
	fd = accept(p->fd, (struct sockaddr *) &peer, &salen);
	if (fd < 0)
		return;

	/* When we're above a certain threshold, reject connects
	 * from anyone but local root */
	if (nsocks >= RPC_SOCKTHRESH
	 && (ntohs(p->peer.sin_port) < 512
	  || ntohs(p->peer.sin_port) > 1023
	  || ntohl(p->peer.sin_addr.s_addr) != INADDR_ANY)) {
		close(fd);
		return;
	}

	pnew = xprt_new(fd, p->rinfo.bufsize, 0);
	if (pnew == NULL)
		return;
	pnew->parent    = p;
	pnew->peer      = peer;
	pnew->startrecv = tcp_startrecv;
	pnew->startxmit = tcp_startxmit;
	pnew->receiver  = tcp_receiver;
	pnew->sender    = tcp_sender;

	pnew->startrecv(pnew);
	return;

fail:	close(fd);
}

static inline void
tcp_nextfrag(SVCXPRT *p)
{
	p->rinfo.startofrecord = 1;
	p->rinfo.want = 4;
	p->rinfo.have = 0;
}

static void
tcp_startrecv(SVCXPRT *p)
{
	p->receiving = 1;
	p->rinfo.size = 0;
	tcp_nextfrag(p);
}

static void
tcp_receiver(SVCXPRT *p)
{
	struct svc_rinfo *r = &p->rinfo;
	u_int32_t	marker, oldsize;
	int		n;

	n = read(p->fd, r->buffer + r->have, r->want);
	if (n <= 0) {
		svc_destroy(p);
		return;
	}

	r->have += n;
	r->want -= n;
	if (r->want != 0)
		return;
	
	/* We have read all we wanted to (for now).
	 * Check what we have got: either a 4 byte record
	 * marker, or a record (fragment).
	 */
	if (!r->startofrecord) {
		/* We've read a record (fragment). 
		 * Are there more to come, or can we execute
		 * the request straightaway? */
		if (r->morefrags) {
			tcp_nextfrag(p);
		} else {
			svc_doit(p);
		}
		return;
	}

	/* We have a 4 byte record marker. Dissect it and
	 * see what's coming.
	 */
	r->startofrecord = r->morefrags = 0;
	r->buffer -= 4;
	memcpy(&marker, r->buffer, 4);
	r->want = ntohl(marker) & 0x7FFFFFFF;
	if (marker & 0x80000000)
		r->morefrags = 1;

	/* Sanity checks */
	oldsize = r->size;
	r->size += r->want;
	if (r->size <= oldsize		/* int overflow or want == 0*/
	 || r->size >= r->bufsize) {
		svc_destroy(p);
		return;
	}
}

static void
tcp_sender(SVCXPRT *p)
{
	struct svc_sinfo *s = &p->sinfo;
	int		n;

	if (s->startofrecord) {
		u_int32_t	marker = htonl(s->size);
		u_int8_t	*data;

		data = (u_int8_t *) &marker;
		n = write(p->fd, data + s->sent, 4 - s->sent);
	} else {
		n = write(p->fd, s->buffer + s->sent, s->size - s->sent);
	}

	if (n <= 0) {
		svc_destroy(p);
		return;
	}
	s->sent += n;

	if (s->startofrecord) {
		if (s->sent == 4) {
			s->startofrecord = 0;
			s->sent = 0;
		}
	}
	if (s->sent == s->size) {
		p->receiving = 1;
		free(s->buffer);
		s->buffer = NULL;
	}
}

static void
tcp_startxmit(SVCXPRT *p, XDR *xdrs)
{
	if (p->sinfo.buffer)
		free(p->sinfo.buffer);

	p->sinfo.buffer = xdrs->buffer;
	p->sinfo.size = xdrs->size;
	p->sinfo.sent = 0;
	p->receiving = 0;

	xdrs->buffer = NULL;
	xdrs->size = 0;
}
