/*
 * Simple but hopefully safer RPC service replacement
 *
 * Copright (C) 2000 <okir@caldera.de>
 */

#include "svc_int.h"
#include "auth_unix.h"
#include <malloc.h>
#include <unistd.h>

/*
 * Create a memory based XDR
 */
static void *
xdrmem_getmem(XDR *xdrs, unsigned int size)
{
	void	*ptr;

	/* round up to a multiple of 4 */
	size = (size + 3) & ~(unsigned int) 3;

	ptr = ((caddr_t) xdrs->x_base) + xdrs->x_pos;
	if (xdrs->x_pos + size >= xdrs->x_size)
		return NULL;
	xdrs->x_pos += size;
	return ptr;
}

void
xdrmem_create(XDR *xdrs, void *base, unsigned int size, int op)
{
	memset(xdrs, 0, sizeof(*xdrs));
	xdrs->x_op = op;
	xdrs->x_base = base;
	xdrs->x_size = size;
	xdrs->getmem = xdrmem_getmem;
}

/*
 * Destroy XDR object
 * Currently a no-op because we assume all buffers
 * are static
 */
void
xdr_destroy(XDR *x)
{
	/* NOP */
}

/*
 * Get current position within stream
 */
unsigned int
xdr_getpos(XDR *x)
{
	return x->x_pos;
}

unsigned int
xdr_getavail(XDR *x)
{
	if (x->x_size < x->x_pos)
		return 0;
	return x->x_size - x->x_pos;
}

/*
 * Free contents of a structure
 */
void
xdr_free(xdrproc_t proc, void *objp)
{
	XDR	xdrs;

	memset(&xdrs, 0, sizeof(xdrs));
	xdrs.x_op = XDR_FREE;
	proc(&xdrs, objp);
}

/*
 * XDR primitives
 */
bool_t
xdr_void(void)
{
	return 1;
}

void *
xdr_inline(XDR *x, unsigned int len)
{
	if (!x->getmem)
		return NULL;
	if (x->x_op != XDR_ENCODE && x->x_op != XDR_DECODE)
		return NULL;
	return x->getmem(x, len);
}

static inline bool_t
xdr_word(XDR *x, void *bp, unsigned int size)
{
	u_int32_t	word;
	void		*ptr;

	if (x->x_op == XDR_FREE)
		return 1;

	if ((ptr = xdr_inline(x, 4)) == NULL)
		return 0;

	switch (x->x_op) {
	case XDR_ENCODE:
		switch (size) {
		case 1:	word = *(u_int8_t *)  bp; break;
		case 2:	word = *(u_int16_t *) bp; break;
		case 4:	word = *(u_int32_t *) bp; break;
		case 8:	word = *(u_int64_t *) bp; break;
		default: return 0;
		}
		*(u_int32_t *) ptr = htonl(word);
		break;
	case XDR_DECODE:
		word = ntohl(*(u_int32_t *) ptr);
		switch (size) {
		case 1: *(u_int8_t *)  bp = word; break;
		case 2: *(u_int16_t *) bp = word; break;
		case 4: *(u_int32_t *) bp = word; break;
		case 8: *(u_int64_t *) bp = word; break;
		default: return 0;
		}
		break;
	default:
		return 0;
	}

	return 1;
}

bool_t
xdr_bool(XDR *x, bool_t *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_enum(XDR *x, int *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_char(XDR *x, char *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_short(XDR *x, short *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_int(XDR *x, int *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_long(XDR *x, long *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_u_char(XDR *x, unsigned char *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_u_short(XDR *x, unsigned short *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_u_int(XDR *x, unsigned int *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_u_long(XDR *x, unsigned long *bp)
{
	return xdr_word(x, bp, sizeof(*bp));
}

bool_t
xdr_float(XDR *x, float *bp)
{
	return 0;
}

bool_t
xdr_double(XDR *x, double *bp)
{
	return 0;
}

/*
 * XDR composites
 */

/*
 * Counted array
 */
bool_t
xdr_array(XDR *x, void **arrayp, unsigned int *sizep,
		unsigned int maxsize, unsigned int elsize,
		xdrproc_t proc)
{
	unsigned int	count;
	caddr_t		array;

	if (!xdr_u_int(x, sizep))
		return 0;

	count = *sizep;
	if (x->x_op == XDR_DECODE) {
		if (count > maxsize)
			return 0;
		if (*arrayp == NULL) {
			*arrayp = (void **) malloc(count * elsize);
			if (*arrayp == NULL)
				return 0;
			memset(*arrayp, 0, count * elsize);
		}
	}

	array = (caddr_t) *arrayp;
	while (count--) {
		if (!proc(x, array))
			return 0;
		array += elsize;
	}

	if (x->x_op == XDR_FREE && array) {
		*arrayp = NULL;
		free(array);
	}

	return 1;
}

/*
 * Fixed length array
 */
bool_t
xdr_vector(XDR *x, void *array, unsigned int size, unsigned int elsize, xdrproc_t proc)
{
	while (size--) {
		if (!proc(x, array))
			return 0;
		(caddr_t) array += elsize;
	}

	return 1;
}

/*
 * Counted bytes
 */
bool_t
xdr_bytes(XDR *x, void **sp, unsigned int *sizep, unsigned int maxsize)
{
	unsigned int	count;
	void		*ptr;

	if (x->x_op == XDR_FREE) {
		if (*sp)
			free(*sp);
		*sizep = 0;
		*sp = NULL;
		return 1;
	}

	if (!xdr_u_int(x, sizep))
		return 0;

	count = *sizep;
	if (count > maxsize || !(ptr = xdr_inline(x, count)))
		return 0;

	switch (x->x_op) {
	case XDR_DECODE:
		if (*sp == NULL && (*sp = (char *) malloc(count + 1)) == NULL)
			return 0;
		memcpy(sp, ptr, count);
		sp[count] = '\0';
		break;

	case XDR_ENCODE:
		memcpy(ptr, sp, count);
		break;
	
	default:
		return 0;
	}

	return 1;
}

/*
 * String with size limit
 */
bool_t
xdr_string(XDR *x, char **sp, unsigned int maxsize)
{
	unsigned int	size;

	if (x->x_op == XDR_ENCODE)
		size = sp? strlen(*sp) : 0;

	return xdr_bytes(x, (void **) sp, &size, maxsize);
}

/*
 * Unlimited string
 */
bool_t
xdr_wrapstring(XDR *x, char **sp)
{
	return xdr_string(x, sp, -1);
}

bool_t
xdr_opaque(XDR *x, void *bp, unsigned int size)
{
	void	*ptr;

	if (x->x_op == XDR_FREE)
		return 1;

	if (!(ptr = xdr_inline(x, size)))
		return 0;
	
	switch (x->x_op) {
	case XDR_DECODE:
		memcpy(bp, ptr, size);
		break;

	case XDR_ENCODE:
		memcpy(ptr, bp, size);
		break;
	
	default:
		return 0;
	}

	return 1;
}

bool_t
xdr_pointer(XDR *x, void **objpp, unsigned int objsize, xdrproc_t proc)
{
	bool_t	haveptr;

	switch (x->x_op) {
	case XDR_ENCODE:
		if (*objpp == NULL) {
			haveptr = 0;
			return xdr_bool(x, &haveptr);
		}
		break;

	case XDR_DECODE:
		if (!xdr_bool(x, &haveptr))
			return 0;
		if (!haveptr)
			return 1;
		break;
	
	default:
		break;
	}

	if (!xdr_reference(x, objpp, objsize, proc))
		return 0;
	
	if (x->x_op == XDR_FREE) {
		free(*objpp);
		*objpp = NULL;
	}

	return 1;
}

bool_t
xdr_reference(XDR *x, void **objpp, unsigned int objsize, xdrproc_t proc)
{
	if (x->x_op == XDR_DECODE) {
		if (*objpp == NULL && (*objpp = malloc(objsize)) == NULL)
			return 0;
		memset(*objpp, 0, objsize);
	}

	return proc(x, *objpp);
}

bool_t
xdr_union(XDR *x, int *dscmp, void *unp, struct xdr_discrim *choice,
		xdrproc_t dfault)
{
	int	dscm;

	if (!xdr_enum(x, dscmp))
		return 0;
	dscm = *dscmp;

	while (choice->proc) {
		if (choice->value == dscm)
			return (*choice->proc)(x, unp);
	}

	if (!dfault)
		return 0;
	return dfault(x, unp);
}

/*
 * Serialize struct opaque_auth
 */
bool_t
xdr_opaque_auth(XDR *x, struct opaque_auth *oa)
{
	return xdr_enum(x, &oa->oa_flavor)
	    && xdr_bytes(x, &oa->oa_base, &oa->oa_length, MAX_AUTH_BYTES);
}

/*
 * rpc_msg
 */
static inline bool_t
xdr_callbody(XDR *x, struct call_body *cb)
{
	return xdr_u_int(x, &cb->cb_rpcvers)
	    && xdr_u_int(x, &cb->cb_prog)
	    && xdr_u_int(x, &cb->cb_vers)
	    && xdr_u_int(x, &cb->cb_proc)
	    && xdr_opaque_auth(x, &cb->cb_cred)
	    && xdr_opaque_auth(x, &cb->cb_verf);
}

bool_t
xdr_callhdr(XDR *x, struct rpc_msg *m)
{
	return xdr_u_int(x, &m->rm_xid)
	    && xdr_enum(x, (int *) &m->rm_direction)
	    && m->rm_direction == CALL
	    && xdr_callbody(x, &m->rm_call);
}

static inline bool_t
xdr_rejectedreply(XDR *x, struct rejected_reply *rj)
{
	if (!xdr_enum(x, (int *) &rj->rj_stat))
		return 0;
	
	switch (rj->rj_stat) {
	case RPC_MISMATCH:
		return xdr_u_int(x, &rj->rj_vers.low)
		    && xdr_u_int(x, &rj->rj_vers.high);
	case AUTH_ERROR:
		return xdr_u_int(x, &rj->rj_why);
	}

	return 1;
}

static inline bool_t
xdr_acceptedreply(XDR *x, struct accepted_reply *ar)
{
	if (!xdr_opaque_auth(x, &ar->ar_verf)
	 || !xdr_enum(x, (int *) &ar->ar_stat))
	 	return 0;
	
	switch (ar->ar_stat) {
	case SUCCESS:
		if (!ar->ar_results.proc)
			return 0;
		return (*ar->ar_results.proc)(x, ar->ar_results.where);

	case PROG_MISMATCH:
		return xdr_u_int(x, &ar->ar_vers.low)
		    && xdr_u_int(x, &ar->ar_vers.high);

	default:
		break;
	}

	return 1;
}

static inline bool_t
xdr_replybody(XDR *x, struct reply_body *rp)
{
	if (!xdr_enum(x, (int *) &rp->rp_stat))
		return 0;
	switch (rp->rp_stat) {
	case MSG_ACCEPTED:
		return xdr_acceptedreply(x, &rp->rp_acpt);
	case MSG_DENIED:
		return xdr_rejectedreply(x, &rp->rp_rjct);
	}

	return 0;
}

bool_t
xdr_replymsg(XDR *x, struct rpc_msg *m)
{
	return xdr_u_int(x, &m->rm_xid)
	    && xdr_enum(x, (int *) &m->rm_direction)
	    && m->rm_direction == REPLY
	    && xdr_replybody(x, &m->rm_reply);
}

/*
 * UNIX credentials
 */
bool_t
xdr_authunix_parms(XDR *x, struct authunix_parms *aup)
{
	return xdr_u_int(x, &aup->aup_time)
	    && xdr_string(x, &aup->aup_machname, AUTH_UNIX_MAX_MACHNAME)
	    && xdr_u_int(x, &aup->aup_uid)
	    && xdr_u_int(x, &aup->aup_gid)
	    && xdr_array(x, (void **) &aup->aup_gids, &aup->aup_len, AUTH_UNIX_NGRPS,
	    			sizeof(aup->aup_gids[0]),
				(xdrproc_t) xdr_u_int);
}
